#pragma once
#include "common/defines.glsl"
#include "common/shadow/shadows.glsl"
#include "common/uniforms/camera.glsl"
#include "common/uniforms/lights.glsl"
#include "common/uniforms/gbuffer.glsl"
#include "common/structs/material.glsl"

float fSpecularIntensity = 5f;

vec3 vEyeDirection;
vec2 vScreenTexCoord;
vec3 vGPos;
vec4 vGColor;
vec3 vGNormal;
Material mMaterial;
MaterialFlags mMaterialFlags;

vec4 calcLightInternal(LightBase light,
                       vec3 lightDirection,
                       float shadowFactor)
{
    vec4 ambientColor = light.vColor * light.fAmbientIntensity;
    float diffuseFactor = max(dot(-lightDirection, vGNormal), 0.0);

    vec4 diffuseColor  = vec4(COLOR_BLACK);
    vec4 specularColor = vec4(COLOR_BLACK);

    diffuseColor = light.vColor * light.fDiffuseIntensity * diffuseFactor;

    vec3 eyeVec = normalize(vEyeDirection);
    vec3 reflectDir = reflect(lightDirection, vGNormal);
    vec3 halfwayDir = normalize(-lightDirection + eyeVec);
    float specularFactor = max(dot(vGNormal, halfwayDir), 0.0);

    specularFactor = pow(specularFactor, mMaterial.fShininess);
    specularColor = light.vColor * (fSpecularIntensity * light.fSpecularIntensity) * specularFactor * mMaterial.vSpecular;

    return (ambientColor + (diffuseColor + specularColor) * shadowFactor);
}

vec4 calcPointLightInternal(PointLightBase pointLightBase,
                            vec3 lightDirection,
                            float shadowFactor,
                            float fragDist) {

    vec4 color = calcLightInternal(pointLightBase.lightBase, lightDirection, shadowFactor);

    float attenuation = 1f;
    if (pointLightBase.bUseAttenuation) {
        attenuation =  pointLightBase.attenuation.fConstant +
                       pointLightBase.attenuation.fLinear * fragDist +
                       pointLightBase.attenuation.fQuadric * fragDist * fragDist;
        attenuation = max(1.0, attenuation);
    }

    return color / attenuation;
}

vec4 calcPointLight(PointLight pointLight)
{
    if (mMaterialFlags.bDisableLighting) return COLOR_BLACK;

    vec3 lightDirection = vGPos - pointLight.pointLightBase.vPosition;

    float shadowFactor = 1;
    #ifdef POINT_LIGHT_SHADOWS
    if (pointLight.pointLightBase.bCastShadows) {
        shadowFactor = calcShadowFactorFromPointLight(vEyeDirection, lightDirection, pointLight.pointLightBase.fShadowDistance);
    }
    #endif

    float dist = length(lightDirection);

    lightDirection = normalize(lightDirection);

    vec4 color = calcPointLightInternal(pointLight.pointLightBase, lightDirection, shadowFactor, dist);
    return color;
}

vec4 calcDirectionalLight(DirectionalLight directionalLight)
{
    if (mMaterialFlags.bDisableLighting) return COLOR_BLACK;
    float shadowFactor = 1;

    #ifdef DIRECTIONAL_LIGHT_SHADOWS
    if (directionalLight.pointLightBase.bCastShadows) {
        vec4 fragPosLightSpace = directionalLight.mLightSpaceTransform * vec4(vGPos, 1);
        shadowFactor = calcShadowFactorFromDirectionalLight(vEyeDirection, fragPosLightSpace, iLightType);
    }
    #endif

    vec3 lightDirection = vGPos - directionalLight.pointLightBase.vPosition;
    float dist = length(lightDirection);

    lightDirection = normalize(lightDirection);

    vec4 color = calcPointLightInternal(directionalLight.pointLightBase,
                                        lightDirection, shadowFactor, dist);
    return color;
}

vec4 calcSpotLight(SpotLight spotLight)
{
    if (mMaterialFlags.bDisableLighting) return COLOR_BLACK;

    float shadowFactor = 1;

    #ifdef POINT_LIGHT_SHADOWS
    if (spotLight.directionalLightBase.pointLightBase.bCastShadows) {
        vec4 fragPosLightSpace = spotLight.directionalLightBase.mLightSpaceTransform * vec4(vGPos, 1);
        shadowFactor = calcShadowFactorFromDirectionalLight(vEyeDirection, fragPosLightSpace, iLightType);
    }
    #endif

    vec3 lightDirection = vGPos - spotLight.directionalLightBase.pointLightBase.vPosition;
    float dist = length(lightDirection);

    lightDirection = normalize(lightDirection);

    vec3 spotDirection = normalize(spotLight.directionalLightBase.vDirection);
    float theta = dot(lightDirection, spotDirection);
    float epsilon = spotLight.fCutOff - spotLight.fOuterCutOff;
    float intensity = clamp((spotLight.fOuterCutOff - theta) / epsilon, 0.0, 1.0);

    if(theta < spotLight.fCutOff)
    {
        intensity = 0;
    }

    vec4 color = intensity * calcPointLightInternal(spotLight.directionalLightBase.pointLightBase,
                                                    lightDirection, shadowFactor, dist);
    return color;
}

vec2 calcScreenTexCoord()
{
    return gl_FragCoord.xy / vViewportSize;
}

void init() {
     vScreenTexCoord = calcScreenTexCoord();
     vGPos = texture(gBuffer.positionTexture, vScreenTexCoord).xyz;
     vGColor = texture(gBuffer.diffuseTexture, vScreenTexCoord);
     vGNormal = texture(gBuffer.normalTexture, vScreenTexCoord).xyz;

     vEyeDirection = camera.vPosition - vGPos;

     vec4 vGMisc = texture(gBuffer.miscTexture, vScreenTexCoord);

     mMaterial.vSpecular = vec4(vGMisc.x, vGMisc.x, vGMisc.x, 1.0);
     mMaterial.fShininess = vGMisc.y;
     mMaterial.iFlags = int(vGMisc.w);

     decodeMaterialFlags(mMaterial.iFlags, mMaterialFlags);
}