#version 330
#include "common/defines.glsl"
#include "common/uniforms/textures.glsl"
#include "common/uniforms/material.glsl"
#include "common/uniforms/camera.glsl"
#include "common/structs/material.glsl"
#include "common/structs/vertex.glsl"
#include "common/locations/gbuffer.glsl"
#include "common/environment/mapping.glsl"
#include "common/debug.glsl"
#include "common/paralax_mapping.glsl"

in Vertex VS_out;

MaterialFlags materialFlags;

void main()
{
    decodeMaterialFlags(material.iFlags, materialFlags);

    vec3 normal = VS_out.vNormal;
    vec3 cameraPos = camera.vPosition;
    vec3 vertexPos = VS_out.vPosition;
    vec2 texCoord = VS_out.vTexCoord;

    // Eat very much GPU ticks!!!
    #ifdef BUMP_MAPPNING
    float dist = distance(VS_out.vPosition, camera.vPosition);
    if (dist <= BUMP_ENABLE_DISTANCE) {

        #ifdef PARALAX_MAPPNING
        if (materialFlags.bUseHeightTexture) {
            vec3 viewDir = VS_out.mInvTBN * (cameraPos - vertexPos);
            texCoord = parallaxMapping(texCoord, normalize(viewDir));
        }
        #endif

        if (materialFlags.bUseNormalTexture) {
            normal = normalize(texture(textures.normal, texCoord).xyz * 2.0 - 1.0);
            normal = VS_out.mTBN * normal;
        }
    }
    #endif

    vec4 diffuse = texture(textures.color, texCoord);
    if (diffuse.a < 0.5) discard;

    diffuse *= material.fDiffuseIntensity;
    diffuse += material.vAmbient * material.fAmbientIntensity;

    #ifdef ENV_MAPPING
    vec3 viewDir = normalize(vertexPos - cameraPos);
    vec4 env;
    if (materialFlags.bEnableEnvReflection) {
        env = calcReflection(viewDir, normal, material);
    }
    if (materialFlags.bEnableEnvRefraction) {
        env = calcRefraction(viewDir, normal, material, env);
    }
    diffuse += env;
    #endif

    vGBufferPositionOut     = vec4(vertexPos, 1.0);
    vGBufferNormalOut       = vec4(normal, 1.0);
    vGBufferMiscOut         = vec4(texture(textures.specular, texCoord).r, material.fShininess, 0, material.iFlags);
    vGBufferDiffuseOut      = diffuse;

    #ifdef DEBUG
    if (isDebug(SHADER_DEBUG_FLAG_GBUFFER_POSITION)) vGBufferDiffuseOut  = vGBufferPositionOut;
    if (isDebug(SHADER_DEBUG_FLAG_GBUFFER_NORMAL)) vGBufferDiffuseOut  = vGBufferNormalOut;
    if (isDebug(SHADER_DEBUG_FLAG_GBUFFER_MISC)) vGBufferDiffuseOut  = vGBufferMiscOut;
    if (isDebug(SHADER_DEBUG_FLAG_TEXTURE_COLOR)) vGBufferDiffuseOut  = texture(textures.color, texCoord);
    if (isDebug(SHADER_DEBUG_FLAG_TEXTURE_NORMAL)) vGBufferDiffuseOut  = texture(textures.normal, texCoord);
    if (isDebug(SHADER_DEBUG_FLAG_TEXTURE_HEIGHT)) vGBufferDiffuseOut  = texture(textures.height, texCoord);
    if (isDebug(SHADER_DEBUG_FLAG_TEXTURE_SPECULAR)) vGBufferDiffuseOut  = texture(textures.specular, texCoord);
    #endif
}
